/*
 * $Id: cards.h,v 1.14 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#define CMD_LINE_BUILD
#define EMULATION_LEVEL	    EMULATION_NONE
#define MEMORY_FORCE_LOCK   MEMORY_FORCE_LOCK_OFF

#define CARD_SECT __attribute__ ((section ("cards")))
#define GAME_SECT __attribute__ ((section ("games")))

#include <PalmOs.h>
#include <stdio.h>

/*
 * Change log
 *
 * $Log: cards.h,v $
 * Revision 1.14  2005/01/11 17:19:00  pdclose
 * Update copyright notice
 *
 * Revision 1.13  2005/01/10 00:49:16  pdclose
 * Make room in the "games" prefs structure for new keys or preferences.  Add a
 * pref to remember the last silk state.
 * 
 * Revision 1.12  2004/12/28 00:31:08  pdclose
 * Add 'games' section for rules
 * 
 * Revision 1.11  2004/12/24 22:24:43  pdclose
 * Change progress numbers: Back renamed to NoHint and changed ProgressSome to
 * a high number so games can use other values to fine-tune hints.
 * 
 * Revision 1.10  2004/04/13 15:37:21  pdclose
 * Moved hintDrawTicks from game struct to global.  There was no need to
 * preserve its value across sessions.
 * 
 * Revision 1.9  2004/04/06 15:59:20  pdclose
 * Copyright date update
 * 
 * Revision 1.8  2004/04/05 22:33:53  pdclose
 * Clarify empty card "reminders"
 * 
 * Revision 1.7  2004/03/25 02:29:40  pdclose
 * Misc changes
 * 
 * Revision 1.6  2003/03/03 23:12:52  pdclose
 * Document EMPTY, NONE, BLANK.
 * Rewrap externs.
 * 
 * Revision 1.5  2002/09/22 23:52:22  pdclose
 * Document cardhint algorithm a bit
 * 
 * Revision 1.4  2002/07/06 21:04:45  pdclose
 * Turn CardDisplay types into an enum, document.
 * 
 * Revision 1.3  2002/07/05 19:05:25  pdclose
 * Move game externs to patience.h, patience.c is the only file that needs them.
 * 
 * Revision 1.2  2002/06/17 01:27:03  pdclose
 * Add new games: Black Widow, Tarantula, and Forty Thieves.
 * 
 * Revision 1.1  2002/04/23 21:54:34  pdclose
 * Initial revision
 */

typedef int		PatI;
typedef unsigned int	PatUI;
typedef long		PatL;
typedef unsigned long	PatUL;
typedef short		PatS;
typedef unsigned short	PatUS;
typedef unsigned char	PatUC;

extern PatUL    hintDrawTicks;

#define MAX_GAMES   32

#define JACK	11
#define QUEEN	12
#define KING	13
#define ACE	14

#define CLUB	    0x00
#define DIAMOND	    0x10
#define HEART	    0x20
#define SPADE	    0x30
#define BACK	    0x40
#define EMPTY	    0x50 // draws an outline, table pattern/color shows through
                         // set to a card rank for a reminder (canfield, etc)
#define NONE	    0x60 // just a placeholder for other cards, no outline
#define BLANK	    0x70 // draws a white-filled outline (usually for stocks)
#define TRUMP	    0x80
#define NextSuit(s) ((s) + 0x10)

#define Suit(c)	    ((c) & 0x70)
#define Trump(c)    ((c) & TRUMP)
#define Rank(c)	    ((c) & 0x0f)
#define MakeCard(t,s,r)	((t) | (s) | (r))

typedef unsigned char	Card;
typedef unsigned char	CardIndex;
typedef unsigned char	CardStackIndex;
typedef unsigned char	CardHistoryKind;
typedef unsigned char	CardBool;

#ifndef True
#define True 1
#define False 0
#endif

// Note: CardDisplayAll, CardDisplaySome, and CardDisplayPack push the
//       layout to the top (or left for horizontal) edge.  Otherwise the
//       layout is centered.
typedef enum {
    CardDisplayTop,	 // Displays just the top card in the pile.
    CardDisplayTopCount, // Displays just the top card in the pile, with a
                         // count of how many cards are in the pile.
    CardDisplayBottom,   // (unknown, displays just the bottom card?)
    CardDisplayAll,      // Displays all cards in the deck.
    CardDisplayAllFixed, // Just like CardDisplayAll, but it doesn't affect
                         // card layout spacing.
    CardDisplayPack,     // Displays all cards in the deck, allows extra
                         // packing on cards that are in 'CurrentPackFunction'
                         // order.
    CardDisplaySome,     // (unknown, not used)
    CardDisplayNone,     // Displays no card; used for the deck when games
                         // don't need a deck.
} CardDisplay;

#define NullIndex	0xff

#define CardFaceUp	1
#define CardFaceDown	0

#define CardStackDeck	0

#define CardShowNone	0
#define CardShowBack	1
#define CardShowPack	2
#define CardShowFace	3

typedef struct {
    CardIndex	    next, prev;
    Card	    card;
    unsigned int    face : 1;
    unsigned int    show : 2;
} CardType;

typedef struct {
    PatS	    x, y;
} PatCoord;

typedef struct {
    PatCoord	    pos;
    CardIndex	    first, last;
    Card	    empty;
    unsigned char   face_up;
    unsigned char   back_up;
    unsigned char   pack_up;
    unsigned int    display : 3;
    unsigned int    dirty : 1;
    unsigned int    horizontal : 1;
    unsigned int    select_card : 1;
    signed char	    face_off;
    signed char	    back_off;
    signed char	    pack_off;
} CardStackType;

extern CardBool (*CurrentPackFunction(void)) (CardIndex, CardIndex);

#define CardHistoryNone	    0
#define CardHistoryMove	    1
#define CardHistoryTurn	    2
#define CardHistoryShuffle  3
#define CardHistoryCallback 4

typedef struct _history {
    PatL    	    serial;
    CardHistoryKind kind;
    union {
	struct {
	    CardStackIndex  stack;
	    CardIndex	    card;
	    unsigned int    face : 1;
	} turn;
	struct {
	    CardStackIndex  source;
	    CardStackIndex  dest;
	    CardIndex	    before;
	    CardIndex	    first;
	    CardIndex	    last;
	} move;
	struct {
	    CardStackIndex  stack;
	    PatL    	    seed;
	} shuffle;
#if 0
    /* doesn't work through application restart */
	struct {
	    void    	    (*func)(PatL closure);
	    PatL    	    closure;
	} callback;
#endif
    } u;
} CardHistoryType;

typedef struct _hint {
    CardStackIndex  stack;
    PatS	    pos;
} CardHintType;

#define MAX_CARDS	104
#define MAX_STACKS	53
#define MAX_ROWCOL	16
#define MAX_HISTORY	1024
#define MAX_HINT	5
#define HISTORY_CHUNK	32
#define HISTORY_CHUNKS	(MAX_HISTORY / HISTORY_CHUNK)

typedef struct _historyChunk {
    CardHistoryType history[HISTORY_CHUNK];
    PatI	    historyTop;
} HistoryChunkType;

typedef struct _gameStat {
    PatUL	    won;
    PatUL	    lost;
} GameStatType;

typedef struct _col 
{
    Boolean         hasArb;
    PatS            numRows;
    Boolean         hasRow[MAX_ROWCOL];
    UInt32          height[MAX_ROWCOL];
} ColumnInfo;

typedef struct _game {
    CardType	    deck[MAX_CARDS];
    CardStackType   stacks[MAX_STACKS];
    CardHintType    hints[MAX_HINT];
    PatS	    nhint;
    CardIndex	    ncards;
    CardStackIndex  nstacks;
    PatL    	    historySerial;
    PatS	    historyTop;
    CardBool	    playing;
    CardBool	    queryMode;
    CardBool	    queryDirty;
    CardBool	    alwaysAuto;
    CardBool	    alwaysDirty;
    CardBool	    leftHanded;
    CardBool	    forceBW;
    CardBool	    useTablePattern;
    CardBool	    alwaysPack;
    CardBool	    centerTopSuit;
    PatUC         empty1;
    PatUC           upActionIdx;
    PatUC         empty2 : 4;
    PatUC         empty2b : 4;
    PatUC         empty2c : 4;
    PatUC           downActionIdx : 4;
    PatUC         empty3 : 4;
    PatUC         empty3b : 4;
    PatUC         empty4 : 4;
    PatUC           jogClickActionIdx : 4;
    PatUC         empty5;
    UInt8           volume;
    PatUC           silkStartNormal : 1;
    PatUC	    gameKind;
    PatS	    gameSeed;
    PatS	    randomSeed;
    CardStackIndex  srcStack;
    CardIndex	    srcCard;
    HistoryChunkType	historyChunk;
    UInt32	    historyID[HISTORY_CHUNKS];
    // This union exists only to avoid changing the size of the game struct,
    // which is stored in preferences.  If the size changes, the struct is
    // cleared, which will reset all the stats!  In the meantime, use the
    // space from unused stats.
    // TODO We wouldn't have this problem if the stats were in a database
    union
    {
        GameStatType    stats[MAX_GAMES];
        struct 
        {
            // WARNING - everying in this struct is cleared if the user
            // clears stats.  GameStatType is two UInt32
            GameStatType spacer[MAX_GAMES-1];
            PatUL           unused1;
            PatUL           unused2;
        } overflow;
    } u;
} GameType;

extern GameType	    game;

#ifdef DEBUG
#define CardPtr(i)	    ((i) == NullIndex ? (ErrDisplay ("CardPtr"), \
						 (CardType *) 0) : &(game.deck[i]))
#define CardStackPtr(i)  ((i) == NullIndex ? (ErrDisplay ("CardStackPtr"), \
						 (CardStackType *) 0) : &(game.stacks[i]))
#define CardHintPtr(i)   ((i) == NullIndex ? (ErrDisplay ("CardHintPtr"), \
						 (CardHintType *) 0) : &(game.hints[i]))
#else
#define CardPtr(i)	    ((i) == NullIndex ? (CardType *) 0 : &(game.deck[i]))
#define CardStackPtr(i)  ((i) == NullIndex ? (CardStackType *) 0 : (&game.stacks[i]))
#define CardHintPtr(i)   ((i) == NullIndex ? (CardHintType *) 0 : &(game.hints[i]))
#endif

typedef struct {
    CardStackIndex  first, last;
} CardStackRangeType;

// CardHint algorithm:
// for each (non-empty) source stack,
//   for each card in the selection range (default = only last card)
//     for each dest stack,
//       record best progress and dest stack
//     if best is not some, try next [could just take the first "some"]
// CardPlay loops through all the stacks in each move, and takes the first
// move with a "some" play, otherwise takes the best play from all moves
//
// Note: CardProgressSome is set high so you can use other values to
// optimize the move if many are valid (see spider.c)
//
typedef enum { 
    CardProgressCant=0, 
    CardProgressNoHint=1, 
    CardProgressNone=2, 
    CardProgressSome=100
} CardProgressType;

typedef CardBool	(*CardSelectStackType) (CardStackIndex stack,
					CardIndex *first, CardIndex *last);
typedef CardProgressType (*CardCheckStackType) (CardStackIndex src, 
				       CardStackIndex dst, CardIndex card);

typedef void	(*CardMoveStackType) (CardStackIndex src,
				      CardStackIndex dst, CardIndex card);

typedef enum { CardMoveNoHelp, CardMoveHelpful, CardMoveAuto, CardMoveHint } CardMoveKind;

typedef struct {
    char		*name;
    CardStackRangeType	hit;
    CardStackRangeType	source;
    CardStackRangeType	dest;
    CardMoveKind	kind;
    CardSelectStackType	SelectStack;
    CardCheckStackType	CheckStack;
    CardMoveStackType	MoveStack;
} CardMoveType;

typedef enum { CardStartDown, CardStartUp, CardStartLastUp } CardStartType;

typedef enum {
    cardStateNone=0, cardStatePacked, cardStateTop, cardStateOther
} CardState;

typedef struct {
    PatCoord		right, left;
    Card		empty;
    PatS    		display;
    CardBool		horizontal;
    CardBool		select_card;
    PatS    		ndeal;
    CardStartType	start;
} CardSetupType;

#define CardEmpty   MakeCard(0, EMPTY, 0)
#define CardNone    MakeCard(0, NONE, 0)
#define CardBlank   MakeCard(0, BLANK, 0)

typedef struct {
    Card		first, last;
    PatS    		ndecks;
    PatS    		nstacks;
    CardBool		(*Normalize) (void);
    void		(*Initialize) (void);
    CardBool		(*Pack) (CardIndex, CardIndex);
    CardSetupType	*setup;
    PatS    		nmoves;
    CardMoveType	*moves;
} RulesType;

extern void Watch (char *);
extern void WatchI (char *, PatL);

void CardSetShoulds (CardStackIndex stack) CARD_SECT;
void CardRedisplay (void) CARD_SECT;
void CardDrawPos (PatS x, PatS y, PatS xlimit, PatS ylimit,
        Card card, CardBool horizontal, CardState cardState) CARD_SECT;
void CardDrawHighlight (CardStackIndex stack, PatI pos);
void CardUndrawHighlight (CardStackIndex stack, PatI pos);
void CardShowHint (CardStackIndex stack, PatI pos) CARD_SECT;
void CardShowMoveHint (CardStackIndex, CardStackIndex, CardIndex, CardIndex) CARD_SECT;
void CardUnshowHints (void) CARD_SECT;
CardBool CardHit (PatI, PatI, CardStackIndex *, CardIndex *, PatI *, PatI *) CARD_SECT;
CardBool CardStartAnimate (PatI x, PatI y, CardStackIndex, CardIndex) CARD_SECT;
void CardStopAnimate (void) CARD_SECT;
void CardMoveAnimate (PatI x, PatI y) CARD_SECT;
void CardShuffle (CardStackIndex stack, CardBool remember) CARD_SECT;
void CardMakeDeck (Card first, Card last, PatI ndecks) CARD_SECT;
void CardInitStack (CardStackIndex stack, PatS x, PatS y, Card empty, PatS display, CardBool horizontal, CardBool select_card) CARD_SECT;
CardBool CardIsInOrder (CardIndex a, CardIndex b) CARD_SECT;
CardBool CardIsInSuitOrder (CardIndex a, CardIndex b) CARD_SECT;
CardBool IsAlternateSuit (Card a, Card b) CARD_SECT;
CardIndex CardInSuitOrder (CardIndex card) CARD_SECT;
CardIndex CardInOrder (CardIndex card) CARD_SECT;
CardIndex CardInReverseSuitOrder (CardIndex card) CARD_SECT;
CardIndex CardInReverseOrder (CardIndex card) CARD_SECT;
CardBool CardIsInAlternatingSuitOrder (CardIndex a, CardIndex b) CARD_SECT;
CardIndex CardInAlternatingSuitOrder (CardIndex card) CARD_SECT;
CardIndex CardInReverseAlternatingSuitOrder (CardIndex card) CARD_SECT;
CardBool CardIsInCircularOrder (CardIndex a, CardIndex b) CARD_SECT;
CardBool CardIsInSuitCircularOrder (CardIndex a, CardIndex b) CARD_SECT;
CardIndex CardInSuitCircularOrder (CardIndex card) CARD_SECT;
CardIndex CardInReverseSuitCircularOrder (CardIndex card) CARD_SECT;
CardBool CardIsInAlternatingSuitCircularOrder (CardIndex a, CardIndex b) CARD_SECT;
CardIndex CardInAlternatingSuitCircularOrder (CardIndex card) CARD_SECT;
CardIndex CardInReverseAlternatingSuitCircularOrder (CardIndex card) CARD_SECT;
void CardTurn (CardStackIndex stack, CardIndex card, CardBool face, CardBool remember) CARD_SECT;
void CardPlay (const RulesType *rules, CardStackIndex src, CardStackIndex dst, CardIndex card) CARD_SECT;
void CardAutoPlay (const RulesType *rules) CARD_SECT;
void CardHint (const RulesType *rules) CARD_SECT;
void CardMoveCard (CardStackIndex from_stack, CardIndex first_card, CardIndex last_card, CardStackIndex to_stack, CardIndex to_card, CardBool remember) CARD_SECT;
void CardMove (CardStackIndex from_stack, CardIndex from_card, CardStackIndex to_stack, CardBool remember) CARD_SECT;
void CardInitHistory (void) CARD_SECT;
#if 0
/* Can't work through application restart */
void	    CardRecordHistoryCallback (void	(*func) (PatL closure),
				       PatL	closure);
#endif
CardBool	    CardUndo (void) CARD_SECT;
PatL	    CardNextHistory (void) CARD_SECT;
void	    CardInit (const RulesType *rules) CARD_SECT;
void
CardSetHanded (const RulesType    *rule,
	       CardBool	    leftHanded) CARD_SECT;
void	    CardFreezeHistory (void) CARD_SECT;

void
DeleteUndoDB (void);
    
void
CardLabelInit (void) CARD_SECT;
    
void
CardLabelFini (void) CARD_SECT;

extern DmOpenRef    undoDB;
