/*
 * $Id: cotillion.c,v 1.5 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"

#define CotillionDeck    0
#define CotillionDiscard 1
// Right area, autofilled from deck
#define CotillionTable   2
#define CotillionNtable  16
#define CotillionLtable  (CotillionTable + CotillionNtable - 1)
// Left table, stacked, no deal
#define CotillionReserve   (CotillionTable + CotillionNtable)
#define CotillionNreserve  4
#define CotillionLreserve  (CotillionReserve + CotillionNreserve - 1)
// A is the base of the first 4, 2 is the base for the last 4
#define CotillionPile    (CotillionReserve + CotillionNreserve)
#define CotillionNpile   8
#define CotillionLpile   (CotillionPile + CotillionNpile - 1)
#define CotillionNstack  (CotillionPile + CotillionNpile)

CardSetupType	CotillionSetup[] = {
    { { 1, 3 }, { 1, 3 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartDown },
    { { 2, 3 }, { 2, 3 }, CardEmpty, CardDisplayTopCount, False, False, 1, CardStartUp },

    // CotillionTable
    { { 6, 0 }, { 6, 0 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 7, 0 }, { 7, 0 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 8, 0 }, { 8, 0 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 9, 0 }, { 9, 0 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },

    { { 6, 1 }, { 6, 1 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 7, 1 }, { 7, 1 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 8, 1 }, { 8, 1 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 9, 1 }, { 9, 1 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },

    { { 6, 2 }, { 6, 2 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 7, 2 }, { 7, 2 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 8, 2 }, { 8, 2 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 9, 2 }, { 9, 2 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },

    { { 6, 3 }, { 6, 3 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 7, 3 }, { 7, 3 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 8, 3 }, { 8, 3 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },
    { { 9, 3 }, { 9, 3 }, CardEmpty, CardDisplayTop, False, False, 1, CardStartUp },

    // CotillionReserve
    { { 0, 0 }, { 0, 0 }, CardNone, CardDisplayAllFixed, False, False, 3, CardStartUp },
    { { 1, 0 }, { 1, 0 }, CardNone, CardDisplayAllFixed, False, False, 3, CardStartUp },
    { { 2, 0 }, { 2, 0 }, CardNone, CardDisplayAllFixed, False, False, 3, CardStartUp },
    { { 3, 0 }, { 3, 0 }, CardNone, CardDisplayAllFixed, False, False, 3, CardStartUp },

    // CotillionPile
    // Ace column
    { { 4, 0 }, { 4, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartUp },
    { { 4, 1 }, { 4, 1 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartUp },
    { { 4, 2 }, { 4, 2 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartUp },
    { { 4, 3 }, { 4, 3 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartUp },
    // Two column
    { { 5, 0 }, { 5, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartUp },
    { { 5, 1 }, { 5, 1 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartUp },
    { { 5, 2 }, { 5, 2 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartUp },
    { { 5, 3 }, { 5, 3 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartUp },
};

static CardStackIndex CotillionCalcDestPile (CardIndex card) GAME_SECT;
static CardProgressType CotillionCheckPile (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static void CotillionMovePile (CardStackIndex src, CardStackIndex dst,
        CardIndex card) GAME_SECT;
static void CotillionMoveDeal (CardStackIndex src, CardStackIndex dst,
        CardIndex last) GAME_SECT;
static CardBool CotillionNormalize (void) GAME_SECT;


// place first cards in suit order for neatness
static CardStackIndex
CotillionCalcDestPile (CardIndex card)
{
    CardStackIndex  dst;
    CardType        *cardp = CardPtr (card);

    if (Rank (cardp->card) > 2)
        return -1;  // Error!

    // calculate which column based on rank
    dst = CotillionPile + (Rank (cardp->card) - 1) * 4;

    // add in the appropriate row based on suit
    switch (Suit(cardp->card))
    {
      case SPADE:	dst += 0; break;
      case HEART:	dst += 1; break;
      case DIAMOND:	dst += 2; break;
      case CLUB:	dst += 3; break;
    }

    return dst;
}

static CardProgressType
CotillionCheckPile (CardStackIndex	src,
	       CardStackIndex	dst,
	       CardIndex	card)
{
    CardType        *cardp;
    CardType        *lastp;
    CardStackIndex  dst_ideal;
    CardStackType   *dp = CardStackPtr (dst);

    if (card == NullIndex)
        return CardProgressCant;

    cardp = CardPtr (card);

    if (! cardp->face)
	return CardProgressCant;

    if (dp->last == NullIndex)
    {
	if (Rank (cardp->card) > 2)
            return CardProgressCant;

        // place first cards in suit order for neatness
        dst_ideal = CotillionCalcDestPile (card);

        if (CardStackPtr (dst_ideal)->first != NullIndex)
            return CardProgressCant;
        if (dst == dst_ideal)
            return CardProgressSome;
        // allow drop on any empty square, but hint the correct square
        return CardProgressNone;
    }
    else
    {
        lastp = CardPtr (dp->last);
        if (Suit(lastp->card) == Suit(cardp->card) &&
                Rank(cardp->card) != Rank(CardPtr(dp->first)->card) &&
                (Rank(lastp->card) + 2 == Rank(cardp->card) ||
                 (Rank(lastp->card) == QUEEN && Rank(cardp->card) == 1) ||
                 (Rank(lastp->card) == KING && Rank(cardp->card) == 2) ))
	    return CardProgressSome;
    }
    return CardProgressCant;
}

static void
CotillionMovePile (CardStackIndex	src,
	       CardStackIndex	dst,
	       CardIndex	card)
{
    CardStackIndex  deck;
    CardIndex       deal;
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
    {
        // place first cards in suit order for neatness
        dst = CotillionCalcDestPile (card);
    }

    CardMove (src, card, dst, True);

    // autofill the table from the discard, or from the deck if
    // the discard is empty
    if (CotillionTable <= src && src <= CotillionLtable)
    {
        deck = CotillionDiscard;
        deal = CardStackPtr (deck)->last;
	if (deal == NullIndex)
        {
            deck = CotillionDeck;
            deal = CardStackPtr (deck)->last;
            if (deal == NullIndex)
                return;
        }
	CardMove (deck, deal, src, True);
	CardTurn (src, CardStackPtr (src)->last, True, True);
    }

    // autodeal the discard, if necessary
    if (CardStackPtr (CotillionDiscard)->last == NullIndex)
        CotillionMoveDeal (CotillionDeck, CotillionDiscard, 0);
}

static void
CotillionMoveDeal (CardStackIndex src,
	      CardStackIndex	dst,
	      CardIndex		last)
{
    CardIndex	card;

    card = CardStackPtr (src)->last;
    if (card != NullIndex)
    {
        CardMove (src, card, dst, True);
        CardTurn (dst, card, True, True);
    }
}

CardMoveType	CotillionMoves[] = {
    /* 0    Move card from reserve to pile */
    {	"reserve to pile",
	{ CotillionReserve, CotillionLreserve },
	{ CotillionReserve, CotillionLreserve },
	{ CotillionPile, CotillionLpile },
	CardMoveAuto,
	0,
	CotillionCheckPile,
	CotillionMovePile
    },
    /* 1    Move card from discard to pile */
    {	"discard to pile",
	{ CotillionDiscard, CotillionDiscard },
	{ CotillionDiscard, CotillionDiscard },
	{ CotillionPile, CotillionLpile },
	CardMoveAuto,
	0,
	CotillionCheckPile,
	CotillionMovePile
    },
    /* 2    Move card from table to pile */
    {	"table to pile",
	{ CotillionTable, CotillionLtable },
	{ CotillionTable, CotillionLtable },
	{ CotillionPile, CotillionLpile },
	CardMoveAuto,
	0,
	CotillionCheckPile,
	CotillionMovePile
    },
    /* 3    deal */
    {	"deal",
	{ CotillionDeck, CotillionDeck },
	{ CotillionDeck, CotillionDeck },
	{ CotillionDiscard, CotillionDiscard },
	CardMoveHelpful,
	0,
	0,
	CotillionMoveDeal,
    },
};

#define NMOVES (sizeof (CotillionMoves) / sizeof (CotillionMoves[0]))

static CardBool
CotillionNormalize (void)
{
    Card	    endrank;
    CardStackIndex  stack;
    CardStackType   *sp;

    for (stack = CotillionPile; stack <= CotillionLpile; stack++)
    {
	sp = CardStackPtr (stack);
	if (sp->last == NullIndex)
	    return False;
        endrank = (stack - CotillionPile) / 4 + QUEEN;
        if (Rank(CardPtr(sp->last)->card) != endrank)
            return False;
    }
    return True;
}

RulesType   cotillionGame = {
    MakeCard (0, CLUB, 1), MakeCard (0, SPADE, KING), 2, CotillionNstack,
    CotillionNormalize,
    0,
    0,
    CotillionSetup,
    NMOVES,
    CotillionMoves
};
