/*
 * $Id: patience.c,v 1.17 2005/01/11 16:55:26 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 *
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 *
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "patience.h"
#include "patienceRsc.h"
#include <System/SysEvtMgr.h>
#include <SonyCLIE.h>
// HandEra
#include <Vga.h>
// Palm 5-way controller
//#include <palmOneChars.h>
#include <palmOneNavigator.h>

/*
 * Change log
 *
 * $Log: patience.c,v $
 * Revision 1.17  2005/01/11 16:55:26  pdclose
 * Fix changelog tags
 *
 * Revision 1.16  2005/01/11 16:35:31  pdclose
 * Update copyright notice for 2005
 * ----------------------------
 * revision 1.15  2005/01/10 02:30:55
 * Return if diamode not set
 * ----------------------------
 * revision 1.14  2005/01/10 01:04:32
 * - Support PalmOS 5.3+ Dynamic Input Area and Pen Input Manager, the
 *   common approach to what Sony calls Silkscreen Manager.
 * - Fix a problem where the main form had a border, which threw off screen
 *   calculations.
 * - Undo any hints if we need to draw something while a hint is still being
 *   displayed.
 * - Implement an alternative to FrmGetObjectIndex, which complained in one
 *   of the later simulators (seems it doesn't like the "try it and see"
 *   approach I was using to handle both stats and new game forms).
 * - Create a non-transparent white for drawing cards (the default
 *   transparency index is 0, which is white).
 * ----------------------------
 * revision 1.13  2004/12/28 00:35:16
 * Replace old entertain routine with new function (GameWonEntertain).
 * ----------------------------
 * revision 1.12  2004/12/24 22:47:17
 * Fix "next game" feature.
 * ----------------------------
 * revision 1.11  2004/04/13 15:37:21
 * Moved hintDrawTicks from game struct to global.  There was no need to
 * preserve its value across sessions.
 * ----------------------------
 * revision 1.10  2004/04/06 15:59:20
 * Copyright date update
 * ----------------------------
 * revision 1.9  2004/04/05 22:43:05
 * Move SetColumnSizes to after key globals are initialized.
 * Fix bug with hint where it would undo a screen update.
 * Call CardDrawRandomLoop on win (scatters random cards around the screen).
 * ----------------------------
 * revision 1.8  2004/03/25 02:21:19
 * Add support for horizontal split screen
 * Fix "Next Game" button on new game screen
 * Make hint time more predictable
 * ----------------------------
 * revision 1.7  2003/11/29 16:27:09
 * Support "game numbers" (seeds)
 * Move Random to cards.c (only used by shuffle)
 * Support Sony silkscreen area
 * ----------------------------
 * revision 1.6  2003/03/03 23:13:32
 * New games: Terrace, Grounds for Divorce, Gypsy
 * Add Sony silkscreen manager mode.
 * Fix for left-handed mode.
 * Remove bitsSaved code that doesn't work.
 * Fix some extraneous refreshes.
 * Change new game logic to allow starting a new game of the same kind.
 * ----------------------------
 * revision 1.5  2002/09/22 23:51:32
 * Don't set hr/qvga mode if the bitmaps aren't present
 * ----------------------------
 * revision 1.4  2002/09/19 19:36:12
 * Change GetObjectPtr to GetObjectPtrById for clarity.
 * Minor rework to consolidate button actions.
 * Preliminary (non-functioning) stats db work.
 * ----------------------------
 * revision 1.3  2002/07/06 21:06:32
 * Add new games, add 'd' to deal a new card.
 * ----------------------------
 * revision 1.2  2002/06/17 01:32:38
 * Add comments for adding a new game.  Add three new games.  Fix bug where
 * the table pattern kept getting reset on monochrome devices.  Split card
 * spacing into X and Y values so horizontally compressed decks don't
 * unnecessarily compress vertically as well.
 * ----------------------------
 * revision 1.1  2002/04/23 21:54:34
 * Initial revision
 *
 */

// To add a new game:
// - Create the rules code (<gamename>.c). Note - Normalize checks the
//   condition for winning)
// - Add gamename.c to Makefile
// - Add the new RulesType to the end of the games array in patience.c
// - Add an extern for the new extern RulesType to patience.h
// - Add game to init in stats.c
// - Add a game button definition to patienceRsc.{rcp,h} (<gamename>Button)
// - Add the name of the game to the name<gamename> defines in
//   patienceRsc.{rcp,h} (rules<gamename>) and add to the list of strings
//   later in patienceRsc.rcp
// - Document the rules in patienceRsc.{rcp,h} (rules<gamename>) and add to
//   the list of strings later in patienceRsc.rcp
// - Adjust newForm in patienceRsc.rcp
//   (may need to adjust row/col in patienceGeom.h)
static const RulesType  *games[] = {
    &klondikeGame,
    &montanaGame,
    &towersGame,
    &spiderGame,
    &tabbyGame,
    &golfGame,
    &acesGame,
    &calcGame,
    &yukonGame,
    &spideretteGame,
    &eightGame,
    &canfieldGame,
    &vegasGame,
    &freecellGame,
    &wishGame,
    &stalactitesGame,
    &tarantulaGame,
    &fortyThievesGame,
    &pictureGame,
    &milliganGame,
    &cotillionGame,
    &terraceGame,
    &bakersDozenGame,
    &gypsyGame,
    //reject pile (adjust list in stats.c if moving games around)
    //&bristolGame, //doesn't play well horizonally
    //&groundsGame, //messy layout
    //&quadrilleGame, //too simple, no challenge at all
};

UInt16 num_games = sizeof (games) / sizeof (games[0]);

GameType	game;
DmOpenRef	undoDB = NULL;
DmOpenRef	statsDB = NULL;
PatUL		hintDrawTicks = 0;

PatS	cardWidth;	/* displayed width of cards */
PatS	cardHeight;	/* displayed height of cards */
PatS	cardSpaceX;	/* space between stacks of cards */
PatS	cardSpaceY;
PatS	cardLabel;	/* offset of label in card */
PatS	cardRadius;	/* corner radius of card */
PatS	cardOriginX;	/* offset of all stacks */
PatS	cardOriginY;
PatS    cardBackOffset;
PatS    cardPackOffset;
PatS	cardLabelWidth;	/* max width of card label element */
PatS	cardLabelHeight;/* max height of card label element */

PatS	xOffset;	/* distance to next card in same stack */
PatS	yOffset;

#define CARD_RADIUS	3
#define CARD_WIDTH	22
#define CARD_HEIGHT	30
PatS maxCardWidth = CARD_WIDTH;
PatS maxCardHeight = CARD_HEIGHT;

static const CustomPatternType  tablePattern = {
    0x11, 0x00, 0x44, 0x00, 0x11, 0x00, 0x44, 0x00
};

static const CustomPatternType  emptyPattern = {
    0, 0, 0, 0, 0, 0, 0, 0
};

// Depends on resource prefsUpActionList, etc.
//LIST "Undo" "Hint" "Examine" "Deal" ID prefs...ActionList
// Note: these are stored in preferences, so don't change the values.
// There is room in the preference settings for 16 different actions
// TODO could use an action that takes the best move, i.e. what would be
// hinted
typedef enum {
    buttonActionUndo=0, buttonActionHint=1, buttonActionExamine=2,
    buttonActionDeal=3, //buttonActionAutoplay=4, buttonActionNewGame=5,
    //buttonActionSameGame=6, buttonActionRestartGame=7,
} ButtonAction;

#define TRACK_TIMEOUT	(sysTicksPerSecond / 5)
#define TRACK_BOX	3
#define HINT_TIMEOUT	(sysTicksPerSecond / 2)

static void StartApplication(void);
static Boolean MainFormHandleEvent(EventPtr event);
static void EventLoop(void);
static void Preferences (void);
static void SetBestColorMode (void);
static void TryHRMode (void);
static void TryHandEraMode (void);
static void TrySilkMode (void);
static void TryDIAMode (void);
static void SizePrefsForm (FormPtr) CARD_SECT;
static void SizeRulesForm (FormPtr) CARD_SECT;
static void HandleAction (ButtonAction);
void DrawTableBackground (RectangleType *r);
static void SelectGame (Boolean, Boolean);
static void SetSizes (void);
static void SetColumnSizes (void);

FormPtr mainFrm;
FormPtr rulesFrm;
Boolean needRedisplay;
CardBool    tracking;
Boolean	    pressed;
Boolean	    awaitingTracking;
PatS	    downX, downY;
UInt32	    downTime;

/*
 * High-res / color support
 * Note that the high-res support changes can be confusing, since form and
 * event coordinates are always 160x160, but window calls can take the full
 * 320x320.
 */
UInt16 hrRefNum;
// High-res mode
HRMode hrMode = noHR;
// HandEra QVGA mode
Boolean qvgaMode = false;
// Sony silk screen support
UInt16 silkRefNum;
Boolean silkMode = false;
Boolean window_resized = false;
// PalmOS Dynamic Input Area support.  See PalmOSCompanion or
// http://www.palmos.com/dev/support/docs/palmos/PalmOSCompanion/PIMConcept.html
// See also:
// Working with Dynamic Input Areas and the Window Manager in Palm OS
// http://kb.palmsource.com/cgi-bin/palmsource.cfg/php/enduser/popup_adp.php?p_faqid=655
Boolean diaMode = false;

UInt32 screen_depth = 0;
UInt32 form_width = 160;
UInt32 form_height = 160;
ColumnInfo column[MAX_ROWCOL];
Boolean supports_color = false;
static UInt32 save_depth = 0;
IndexedColorType tableIndex, cardBackIndex, blackIndex, whiteIndex, packIndex,
        hintIndex;
// Define OS versions so we do the right thing everywhere
// Note that you should use < or >= in these comparisons, in nearly
// every case, because 3.0.1 should be treated the same way as 3.0.
// If you say "if vers <= 3.0" you'd often do the wrong thing.
UInt32  romVersion, sysVersion30, sysVersion33, sysVersion35,
        sysVersion40, sysVersion50;


static void
SetSizes (void)
{
    PatS    	    maxx, maxy;
    CardStackIndex  stack;
    CardStackType   *sp = NULL;
    Boolean	    hasArbX, hasArbY;

    maxx = 0; maxy = 0;
    hasArbX = false;
    hasArbY = false;
    for (stack = 0; stack < game.nstacks; stack++)
    {
	sp = CardStackPtr (stack);
	if (sp->pos.x > maxx)
	    maxx = sp->pos.x;
	if (sp->pos.y > maxy)
	    maxy = sp->pos.y;
	switch (sp->display) {
	case CardDisplayAll:
	case CardDisplaySome:
	case CardDisplayPack:
	    if (sp->horizontal)
		hasArbX = true;
	    else
		hasArbY = true;
	    break;
	default:
	}
    }
    maxx = maxx + 1;
    maxy = maxy + 1;
    if (qvgaMode)
    {
        maxCardWidth = CARD_WIDTH*1.5;
        maxCardHeight = CARD_HEIGHT*1.5;
    }

    // round up (9 cards looks better slightly too wide)
    cardWidth = (form_width + maxx/2) / maxx;
    if (cardWidth > maxCardWidth)
	cardWidth = maxCardWidth;
    cardHeight = (form_height + maxy/2) / maxy;
    if (cardHeight > maxCardHeight)
	cardHeight = maxCardHeight;

    if (cardWidth < maxCardWidth)
    {
	cardSpaceX = 4 - (maxCardWidth - cardWidth);
	if (cardSpaceX < 1)
	    cardSpaceX =  1;
	//cardRadius = 0;
	cardRadius = CARD_RADIUS;
	if (cardSpaceX >= 2)
	    cardLabel = 2;
	else
	    cardLabel = 1;
    }
    else
    {
	cardSpaceX = 4;
	//cardRadius = 1;
	cardRadius = CARD_RADIUS;
	cardLabel = 3;
    }
    if (cardHeight < maxCardHeight)
    {
	cardSpaceY = 4 - (maxCardHeight - cardHeight);
	if (cardSpaceY < 1)
	    cardSpaceY = 1;
    }
    else
	cardSpaceY = 4;

    cardOriginX = 0;
    if (!hasArbX)
	cardOriginX = (form_width - cardWidth * maxx) >> 1;
    cardOriginY = 0;
    if (!hasArbY)
	cardOriginY = (form_height - cardHeight * maxy) >> 1;

    SetColumnSizes ();

    if (hrMode)
    {
        xOffset = cardLabelWidth/2 + cardSpaceX/2 + 1;
        yOffset = cardLabelHeight/2 + cardSpaceY/2 + 3;
    }
    else
    {
        xOffset = cardLabelWidth + cardSpaceX/2 + 2;
        yOffset = cardLabelHeight + cardSpaceY/2 + 1;
    }

    cardBackOffset = 3;
    cardPackOffset = 2;
    if (qvgaMode)
    {
        cardBackOffset = 5;
        cardPackOffset = 3;
    }
}

void
DrawTableBackground (RectangleType *r)
{
    IndexedColorType currBackColor;

    if (screen_depth == 1 || romVersion < sysVersion30)
    {
        // black and white, and/or old OS: use table pattern
        if (game.useTablePattern)
            WinSetPattern (&tablePattern);
        else
            WinSetPattern (&emptyPattern);
        WinFillRectangle (r, 0);
        WinSetPattern (&emptyPattern);
    }
    else if (romVersion < sysVersion35)
    {
        int i;
	RGBColorType rgb, oldrgb;
        if (game.useTablePattern)
        {
            WinSetPattern (&tablePattern);
            WinFillRectangle (r, 0);
            WinSetPattern (&emptyPattern);
        }
        else
        {
            // use the 3.0-3.3 APIs just to draw a gray background
            // WinDrawRectangle didn't obey the color, so we have to draw
            // lines...
            MemSet(&rgb, sizeof(rgb), 0);
            // dark gray is at 0x55, light gray at 0xAA.
            rgb.r = 0xaa; rgb.g = 0xaa; rgb.b = 0xaa;
            WinSetColors(&rgb, &oldrgb, 0, 0);
            for (i=0; i < r->extent.y; i++)
            {
                WinDrawLine(r->topLeft.x, r->topLeft.y+i,
                        (r->topLeft.x + r->extent.x-1), r->topLeft.y+i);
            }
            WinSetColors(&oldrgb, 0, 0, 0);
        }
    }
    else
    {
        // we're not in black and white, and are on a 3.5+ machine;
        // use the good APIs.

        if (supports_color)
            currBackColor = WinSetBackColor(tableIndex);
        else
            currBackColor = WinSetBackColor(packIndex);
        if (game.useTablePattern)
        {
            WinSetPattern (&tablePattern);
            WinFillRectangle (r, 0);
            WinSetPattern (&emptyPattern);
        }
        else
            WinEraseRectangle (r, 0);
        WinSetBackColor(currBackColor);
    }
}

static void
Redisplay (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    RectangleType   r;

    // Undo any hints before any drawing
    if (game.nhint > 0)
        CardUnshowHints ();

    //FrmEraseForm (mainFrm); /* Causes big problems in PalmOS 3.5 */
    if (qvgaMode)
        FntSetFont (VgaBaseToVgaFont (stdFont));
    else
        FntSetFont (stdFont);
    FrmGetFormBounds (mainFrm, &r);
    DrawTableBackground (&r);

    if (game.playing)
    {
	game.queryDirty = True;
	game.alwaysDirty = True;
	for (stack = 0; stack < game.nstacks; stack++)
	{
	    sp = CardStackPtr (stack);
	    if (sp->display != CardDisplayNone)
		sp->dirty = True;
	}
	CardRedisplay ();
    }
}

static void
SetQueryMode (CardBool queryMode)
{
    game.queryMode = queryMode;
    game.queryDirty = True;
    CardRedisplay ();
}

static void
SetAutoMode (CardBool autoMode)
{
    game.alwaysAuto = autoMode;
    game.alwaysDirty = True;
    if (game.alwaysAuto)
	CardAutoPlay (games[game.gameKind]);
    CardRedisplay ();
}

static void
SetHanded (CardBool leftHanded, CardBool redisplay)
{
    if (game.leftHanded == leftHanded)
	return;
    if (game.playing) {
	CardSetHanded (games[game.gameKind], leftHanded);
	if (redisplay)
	    Redisplay ();
    } else {
	game.leftHanded = leftHanded;
    }
}

CardBool
(*CurrentPackFunction (void)) (CardIndex, CardIndex)
{
    return games[game.gameKind]->Pack;
}

FormPtr	dialogFrm;
void    (*DialogIdleFunc)(void);
Boolean	(*DoDialogHandleEvent)(EventPtr);

static Boolean
DialogHandleEvent (EventPtr event)
{
    Boolean ret;

    if (DoDialogHandleEvent)
	ret = (*DoDialogHandleEvent) (event);
    else
	ret = false;
    if (DialogIdleFunc)
    {
	/*
	 * When the form comes up, queue a magic
	 * update event so that when the queue is
	 * mostly empty, we'll get it.  The singing
	 * function will eat any non pen-down events
	 * so we don't want to lose anything important.
	 */
	switch (event->eType) {
	case winEnterEvent:
	case frmOpenEvent:
	    FrmUpdateForm (FrmGetFormId (dialogFrm), 1);
	    break;
	case frmUpdateEvent:
	    if (event->data.frmUpdate.updateCode == 1)
	    {
		ret = true;
		(*DialogIdleFunc) ();
		DialogIdleFunc = 0;
	    }
	    break;
	default:
	}
    }
    return ret;
}

// Check for columns with multiple "arb" stacks, if found, set height
// for that column.
//     example: Baker's Game (but Picture Gallery is a different case)
// TODO: same for rows, may not be worth it
// NOTE: relies on form_height and cardHeight globals
static void
SetColumnSizes (void)
{
    CardStackType   *sp;
    CardStackIndex  stack;
    PatS            col, row, next_row;

    for (col = 0; col < MAX_ROWCOL; col++)
    {
        column[col].hasArb = false;
        column[col].numRows = 0;
        for (row = 0; row < MAX_ROWCOL; row++)
        {
            column[col].hasRow[row] = false;
            column[col].height[row] = form_height;
        }
    }

    // First pass: determine which columns have arbitrary stacks, and
    // which rows are valid for that stack
    for (stack = 0; stack < game.nstacks; stack++)
    {
        sp = CardStackPtr (stack);
        col = sp->pos.x;
        switch (sp->display) {
        case CardDisplayAll:
        case CardDisplaySome:
        case CardDisplayPack:
            if (!sp->horizontal) {
                column[col].hasArb = true;
                column[col].hasRow[sp->pos.y] = true;
                column[col].numRows++;
            }
            break;
        default:
        }
    }

    // Second pass: for each arb column, figure out the height for each row

    // TODO: Could evenly distribute form_height, assume y is ordinal to
    // split, depending on numRows.  Need special setting to avoid
    // misinterpreting.  Have to adjust the stack position as a result.
    for (col = 0; col < MAX_ROWCOL; col++)
    {
        if (column[col].hasArb) {
            for (row = 0; row < MAX_ROWCOL; row++)
            {
                if (column[col].hasRow[row])
                {
                    // find next row
                    for (next_row = row+1; next_row < MAX_ROWCOL; next_row++)
                        if (column[col].hasRow[next_row])
                            break;
                    if (next_row == MAX_ROWCOL)
                        // no next row
                        column[col].height[row] = form_height;
                    else
                        // our height = that row's base
                        column[col].height[row] =
                                next_row * cardHeight + cardOriginY;
                }
            }
        }
    }
}

// HandEra and Sony silkscreen
void
SizeFormWindow (FormPtr frmP, Int16 *dxP, Int16 *dyP)
{
    RectangleType r, old;
    WinHandle formWin = FrmGetWindowHandle (frmP);
    Boolean resize = false;

    if (! qvgaMode && ! silkMode && ! diaMode)
        return;

    if (qvgaMode)
        VgaFormModify (frmP, vgaFormModify160To240);

    if (romVersion >= sysVersion40)
    {
        WinGetBounds(formWin, &old);
        WinGetBounds(WinGetDisplayWindow(), &r);

        if (dxP != NULL)
            *dxP = r.extent.x - old.extent.x;
        if (dyP != NULL)
            *dyP = r.extent.y - old.extent.y;
        if (r.extent.x != old.extent.x || r.extent.y != old.extent.y)
            resize = true;
    }
    else
    {
        r.topLeft.x = 0;
        r.topLeft.y = 0;
        WinGetDisplayExtent(&r.extent.x, &r.extent.y);

        if (dxP != NULL)
            *dxP = r.extent.x - form_width;
        if (dyP != NULL)
            *dyP = r.extent.y - form_height;
        if (r.extent.x != form_width || r.extent.y != form_height)
            resize = true;
    }

    if (resize)
    {
        form_width = r.extent.x;
        form_height = r.extent.y;
        WinSetWindowBounds(formWin, &r);
        SetSizes ();
        needRedisplay = true;
    }
}

void
AdjustObject (FormPtr formP, UInt16 objID,
             Int16 left_delta, Int16 top_delta,
             Int16 width_delta, Int16 height_delta)
{
    RectangleType   r;
    UInt16          formObjIndex = frmInvalidObjectId;
    UInt16          i, numObjects;

    //formObjIndex = FrmGetObjectIndex (formP, objID);
    // alternative to FrmGetObjectIndex, which complains in latest simulator
    numObjects = FrmGetNumberOfObjects (formP);
    for (i = 0; i < numObjects; i++) {
        if (FrmGetObjectId (formP, i) == objID) {
            // found it
            formObjIndex = i;
            break;
        }
    }
    if (formObjIndex == frmInvalidObjectId)
        return;
    FrmGetObjectBounds (formP, formObjIndex, &r);
    r.topLeft.x += left_delta;
    r.topLeft.y += top_delta;
    r.extent.x  += width_delta;
    r.extent.y  += height_delta;
    FrmSetObjectBounds (formP, formObjIndex, &r);
}

Int16
DoDialog (FormPtr   frm,
	  Boolean   (*HandleEvent)(EventPtr))
{
    Int16	    r;

    DoDialogHandleEvent = HandleEvent;
    dialogFrm = frm;
    FrmSetEventHandler (frm, DialogHandleEvent);
    r = FrmDoDialog (frm);
    return r;
}

void
Help (Int16   id)
{
    FormPtr	frm;

    frm = FrmInitForm (id);
    SizeFormWindow (frm, NULL, NULL);
    if (frm)
    {
	(void) DoDialog (frm, 0);
	FrmDeleteForm (frm);
    }
}

static void
SizePrefsForm (FormPtr prefsFrm)
{
    RectangleType r;
    UInt16 width;
    Int16 offset;

    SizeFormWindow (prefsFrm, NULL, NULL);

    // Find max width of up/down/jog labels
    GetObjectBounds (prefsFrm, prefsUpLabel, &r);
    width = r.extent.x;
    GetObjectBounds (prefsFrm, prefsDownLabel, &r);
    if (width < r.extent.x)
        width = r.extent.x;
    GetObjectBounds (prefsFrm, prefsJogClickLabel, &r);
    if (width < r.extent.x)
        width = r.extent.x;
    GetObjectBounds (prefsFrm, prefsVolumeLabel, &r);
    if (width < r.extent.x)
        width = r.extent.x;

    // adjust to right-justify the labels
    GetObjectBounds (prefsFrm, prefsUpLabel, &r);
    offset = width - r.extent.x;
    AdjustObject (prefsFrm, prefsUpLabel, offset, 0, 0, 0);
    AdjustObject (prefsFrm, prefsUpBtnAction, offset, 0, 0, 0);
    AdjustObject (prefsFrm, prefsUpActionList, offset, 0, 0, 0);

    GetObjectBounds (prefsFrm, prefsDownLabel, &r);
    offset = width - r.extent.x;
    AdjustObject (prefsFrm, prefsDownLabel, offset, 0, 0, 0);
    AdjustObject (prefsFrm, prefsDownBtnAction, offset, 0, 0, 0);
    AdjustObject (prefsFrm, prefsDownActionList, offset, 0, 0, 0);

    GetObjectBounds (prefsFrm, prefsJogClickLabel, &r);
    offset = width - r.extent.x;
    AdjustObject (prefsFrm, prefsJogClickLabel, offset, 0, 0, 0);
    AdjustObject (prefsFrm, prefsJogClickBtnAction, offset, 0, 0, 0);
    AdjustObject (prefsFrm, prefsJogClickActionList, offset, 0, 0, 0);

    GetObjectBounds (prefsFrm, prefsVolumeLabel, &r);
    offset = width - r.extent.x;
    AdjustObject (prefsFrm, prefsVolumeLabel, offset, 0, 0, 0);
    AdjustObject (prefsFrm, prefsVolume, offset, 0, 0, 0);
    AdjustObject (prefsFrm, prefsVolumeList, offset, 0, 0, 0);
}

static void
Preferences (void)
{
    UInt16	btn;
    ControlPtr	forceBWCheck;
    ControlPtr	leftHandedCheck;
    ControlPtr	tablePatternCheck;
    ControlPtr	highlightBuildsCheck;
    ControlPtr	centerSuitCheck;
    ControlPtr	upBtnAction;
    ListPtr     upActionList;
    ControlPtr	downBtnAction;
    ListPtr     downActionList;
    ControlPtr	jogClickBtnAction;
    ListPtr     jogClickActionList;
    ListPtr	volumeList;
    ControlPtr	volumeButton;
    FormPtr	prefsFrm = FrmInitForm (prefsForm);

    if (prefsFrm)
    {
        SizePrefsForm (prefsFrm);

        forceBWCheck = GetObjectPtrById (prefsFrm, prefsForceBW);
        if (romVersion < sysVersion30)
            // black and white, and/or old OS, B&W is the only choice
            CtlSetUsable (forceBWCheck, false);
        else
            CtlSetValue (forceBWCheck, game.forceBW);

        leftHandedCheck = GetObjectPtrById (prefsFrm, prefsLeftHanded);
        CtlSetValue (leftHandedCheck, game.leftHanded);

        tablePatternCheck = GetObjectPtrById (prefsFrm, prefsUseTablePat);
        CtlSetValue (tablePatternCheck, game.useTablePattern);

        highlightBuildsCheck = GetObjectPtrById (prefsFrm, prefsAlwaysPack);
        CtlSetValue (highlightBuildsCheck, game.alwaysPack);

        centerSuitCheck = GetObjectPtrById (prefsFrm, prefsCenterTopSuit);
        CtlSetValue (centerSuitCheck, game.centerTopSuit);

        upActionList = GetObjectPtrById (prefsFrm, prefsUpActionList);
        LstSetSelection (upActionList, game.upActionIdx);

        upBtnAction = GetObjectPtrById (prefsFrm, prefsUpBtnAction);
        CtlSetLabel (upBtnAction, LstGetSelectionText (upActionList,
                LstGetSelection(upActionList)));

        downActionList = GetObjectPtrById (prefsFrm, prefsDownActionList);
        LstSetSelection (downActionList, game.downActionIdx);

        downBtnAction = GetObjectPtrById (prefsFrm, prefsDownBtnAction);
        CtlSetLabel (downBtnAction, LstGetSelectionText (downActionList,
                LstGetSelection(downActionList)));

        jogClickActionList =
                GetObjectPtrById (prefsFrm, prefsJogClickActionList);
        LstSetSelection (jogClickActionList, game.jogClickActionIdx);

        jogClickBtnAction =
                GetObjectPtrById (prefsFrm, prefsJogClickBtnAction);
        CtlSetLabel (jogClickBtnAction, LstGetSelectionText (jogClickActionList,
                LstGetSelection(jogClickActionList)));

        volumeList = GetObjectPtrById (prefsFrm, prefsVolumeList);
        if (game.volume == 0)
            LstSetSelection (volumeList, 0);
        else if (game.volume < sndMaxAmp*0.25)
            LstSetSelection (volumeList, 1);
        else if (game.volume < sndMaxAmp*0.75)
            LstSetSelection (volumeList, 2);
        else
            LstSetSelection (volumeList, 3);

        volumeButton = GetObjectPtrById (prefsFrm, prefsVolume);
        CtlSetLabel (volumeButton, LstGetSelectionText (volumeList,
                LstGetSelection(volumeList)));

        for (;;)
        {
            btn = DoDialog (prefsFrm, 0);
            switch (btn)
            {
            case prefsOk:
		// [NJ] Ideally, this would only be set if something changed
		// that required a redisplay.  But for now, always force a
		// redisplay if OK was pressed.  This is now necessary because
		// the extraneous refresh was taken out of DoDialog().
		needRedisplay = True;

                if (game.forceBW != CtlGetValue (forceBWCheck))
                {
                    game.forceBW = CtlGetValue (forceBWCheck);
                    SetBestColorMode ();
                }
                SetHanded (CtlGetValue (leftHandedCheck), false);

                game.useTablePattern = CtlGetValue (tablePatternCheck);

                game.alwaysPack = CtlGetValue (highlightBuildsCheck);

                game.centerTopSuit = CtlGetValue (centerSuitCheck);

                game.upActionIdx = LstGetSelection (upActionList);
                game.downActionIdx = LstGetSelection (downActionList);
                game.jogClickActionIdx = LstGetSelection (jogClickActionList);

                // System settings Low=8, Med=32, High=64 (sndMaxAmp)
                switch (LstGetSelection (volumeList))
                {
                case 0:
                    game.volume = 0;
                    break;
                case 1:
                    game.volume = sndMaxAmp * 0.125;
                    break;
                case 2:
                    game.volume = sndMaxAmp * 0.5;
                    break;
                case 3:
                default:
                    game.volume = sndMaxAmp;
                    break;
                }

                break;
            case prefsCancel:
                // nothing
                break;
            default:
                break;
            }
            break;
        }
	FrmDeleteForm (prefsFrm);
    }
}

static PatS
GameInfoForm (void)
{
    UInt16	btn;
    FieldPtr    fldGameNumber;
    FormPtr	gameInfoFrm = FrmInitForm (gameInfoForm);
    MemHandle   numberH, oldH;
    MemPtr      number_text;
    PatS	gameSeed = -1;

    if (gameInfoFrm)
    {
        //SizePrefsForm (gameInfoFrm);

        fldGameNumber = GetObjectPtrById (gameInfoFrm, gameInfoNumber);

        numberH = MemHandleNew (64);
        number_text = MemHandleLock (numberH);
        StrIToA (number_text, game.gameSeed);
        MemHandleUnlock (numberH);

        oldH = FldGetTextHandle(fldGameNumber);
        FldSetTextHandle(fldGameNumber, numberH);
        FldDrawField(fldGameNumber);
        if (oldH != NULL)
            MemHandleFree(oldH);

        for (;;)
        {
            btn = DoDialog (gameInfoFrm, 0);
            switch (btn)
            {
            case gameInfoOk:
                if (FldDirty (fldGameNumber))
                {
                    number_text = FldGetTextPtr(fldGameNumber);
                    gameSeed = StrAToI (number_text);
                }
                break;
            case gameInfoCancel:
                // nothing
                break;
            default:
                break;
            }
            break;
        }

        needRedisplay = True;
	FrmDeleteForm (gameInfoFrm);
    }
    return gameSeed;
}

static void
RulesSetScrolling (void)
{
    FieldPtr	    field;

    field = GetObjectPtrById (rulesFrm, rulesText);
    FrmUpdateScrollers (rulesFrm,
			FrmGetObjectIndex (rulesFrm, rulesScrollUp),
			FrmGetObjectIndex (rulesFrm, rulesScrollDown),
			FldScrollable (field, winUp),
			FldScrollable (field, winDown));
}

static void
RulesScroll (WinDirectionType dir, CardBool screen)
{
    FieldPtr	    field;
    UInt16	    fieldIndex;
    UInt16	    lines;

    fieldIndex = FrmGetObjectIndex (rulesFrm, rulesText);
    field = FrmGetObjectPtr (rulesFrm, fieldIndex);
    if (screen)
    {
	lines = FldGetVisibleLines (field) - 2;
	if (lines < 1)
	    lines = 1;
    }
    else
	lines = 1;
    FldScrollField (field, lines, dir);
    RulesSetScrolling ();
}

static Boolean
RulesFormHandleEvent(EventPtr event)
{
    Boolean handled = false;

    switch (event->eType) {
    case ctlRepeatEvent:
	switch (event->data.ctlRepeat.controlID) {
	case rulesScrollUp:
	    RulesScroll (winUp, False);
	    break;
	case rulesScrollDown:
	    RulesScroll (winDown, False);
	    break;
	}
	break;
    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case rulesScrollUp:
	    RulesScroll (winUp, False);
	    break;
	case rulesScrollDown:
	    RulesScroll (winDown, False);
	    break;
	}
	break;
    case keyDownEvent:
	switch (event->data.keyDown.chr) {
	case pageDownChr:
	    RulesScroll (winDown, True);
	    handled = true;
	    break;
	case pageUpChr:
	    RulesScroll (winUp, True);
	    handled = true;
	    break;
	}
	break;
    case frmUpdateEvent:
        if (window_resized)
        {
            SizeRulesForm (FrmGetActiveForm ());
            window_resized = false;
        }
        // don't handle to allow system redraw
        break;
    // PalmOS dynamic input area
    case winEnterEvent:
        if (diaMode) {
            // PIM 1.0 compatility
            PINSetInputTriggerState(pinInputTriggerEnabled);
            PINSetInputAreaState(pinInputAreaUser);
        }
        // don't think we need to handle this, we didn't really do anything
        break;
    // PalmOS dynamic input area
    case winDisplayChangedEvent:
        window_resized = true;
        // not needed?
        //FrmUpdateForm (FrmGetActiveFormID (), frmRedrawUpdateCode);
        handled = true;
        break;
    // HandEra
    case displayExtentChangedEvent:
        window_resized = true;
        FrmUpdateForm (FrmGetActiveFormID (), frmRedrawUpdateCode);
        handled = true;
        break;
    default:
    }
    return handled;
}

static void
SizeRulesForm (FormPtr rulesFrm)
{
    Int16           heightDelta;

    if (qvgaMode || silkMode || diaMode)
    {
        SizeFormWindow (rulesFrm, NULL, &heightDelta);
        // rulesText - adjust height
        AdjustObject (rulesFrm, rulesText, 0, 0, 0, heightDelta);
        // rulesOk, rulesScrollUp, rulesScrollDown - adjust y
        AdjustObject (rulesFrm, rulesOk, 0, heightDelta, 0, 0);
        AdjustObject (rulesFrm, rulesScrollUp, 0, heightDelta, 0, 0);
        AdjustObject (rulesFrm, rulesScrollDown, 0, heightDelta, 0, 0);

        RulesSetScrolling ();
    }
}

void
Rules (Int16 id)
{
    FieldPtr	field;
    MemHandle	rulesH;

    rulesH = DmGetResource ('tSTR', id);
    if (rulesH)
    {
	rulesFrm = FrmInitForm (rulesForm);
        SizeRulesForm (rulesFrm);
	field = GetObjectPtrById (rulesFrm, rulesText);
	FldSetInsPtPosition (field, 0);
	FldSetTextHandle (field, rulesH);
	RulesSetScrolling ();
	DoDialog (rulesFrm, RulesFormHandleEvent);
	FldSetTextHandle (field, 0);
	DmReleaseResource (rulesH);
	FrmDeleteForm (rulesFrm);
        rulesFrm = NULL;
    }
}

void
RestartGame (void)
{
    if (!game.playing)
	return;
    CardInit (games[game.gameKind]);
    SetSizes ();
    needRedisplay = True;
}

static Int16
WhichGame (FormPtr newFrm)
{
    UInt16	cur;
    ControlPtr	gameBtn;

    for (cur = 0; cur < num_games; cur++)
    {
	gameBtn = GetObjectPtrById (newFrm, newFormFirstGame + cur);
        if (gameBtn == (void*)-1)
            continue;
	if (CtlGetValue (gameBtn))
	    return cur;
    }
    return -1;
}

void
SelectGameEntertain (void)
{
    PlayRandomSong (True);
}

// gameSeed is a reference to a new random seed.
// Override it if a specific game seed is desired
static UInt16
SelectNewGame (Boolean entertain, PatS *gameSeed)
{
    FormPtr	newFrm;
    ControlPtr	gameBtn;
    ControlPtr	cancelBtn;
    ControlPtr	nextBtn;
    UInt16	btn;
    UInt16	cur;
    UInt16	gameKind = -1;
    PatS	tmpSeed = -1;

    newFrm = FrmInitForm (newForm);
    SizeFormWindow (newFrm, NULL, NULL);
    /*
     * Set game selection buttons appropriately
     */
    for (cur = 0; cur < num_games; cur++)
    {
        gameBtn = GetObjectPtrById (newFrm, newFormFirstGame + cur);
        if (gameBtn == (void*)-1)
            continue;
        if (game.gameKind == cur)
            CtlSetValue (gameBtn, 1);
        else
            CtlSetValue (gameBtn, 0);
    }

    /*
     * Only enable the cancel button if a game is in progress
     * (or the "Next" button if not in progress)
     */
    cancelBtn = GetObjectPtrById(newFrm, newFormCancelBtn);
    CtlSetUsable (cancelBtn, game.playing);

    nextBtn = GetObjectPtrById(newFrm, newFormNextGame);
    CtlSetUsable (nextBtn, !game.playing);

    if (romVersion < sysVersion30 && entertain)
        DialogIdleFunc = SelectGameEntertain;

    for (;;)
    {
        btn = DoDialog (newFrm, 0);
        gameKind = WhichGame (newFrm);
        switch (btn)
        {
        case newFormNewBtn:
            // use provided gameSeed
            break;
        case newFormNextGame:
            tmpSeed = game.gameSeed + 1;
            break;
        case newFormSelectBtn:
            tmpSeed = GameInfoForm ();
            if (tmpSeed == -1 /*cancel*/)
                continue;
            break;
        case newFormRulesBtn:
            if (0 <= gameKind && gameKind < num_games)
                Rules (rulesFirst + gameKind);
            continue;
        case newFormCancelBtn:
            gameKind = -1;
            break;
        default:
            break;
        }
        break;
    }
    if (0 <= gameKind && gameKind < num_games && tmpSeed != -1)
        *gameSeed = tmpSeed;
    FrmDeleteForm (newFrm);
    return gameKind;
}

static void
SelectGame (Boolean entertain, Boolean repeatSameKind)
{
    UInt16	gameKind = -1;
    UInt16	prevGameKind = game.gameKind;
    PatS	gameSeed = TimGetTicks ();

    if (repeatSameKind)
    {
        gameKind = prevGameKind;
    }
    else
    {
        gameKind = SelectNewGame (entertain, &gameSeed);
    }

    if (0 <= gameKind && gameKind < num_games)
    {
        // record stats for previous game
        if (prevGameKind >= 0 && prevGameKind < num_games)
        {
            if (game.playing)
                game.u.stats[prevGameKind].lost++;
            else
                game.u.stats[prevGameKind].won++;
        }
	game.gameKind = gameKind;
	game.gameSeed = gameSeed;
	game.playing = True;
	game.queryMode = False;
	RestartGame ();
    }
}

static void
Save (void)
{
    PrefSetAppPreferencesV10 (PatienceFileCreator, PatienceVersionNum,
			      &game, sizeof (GameType));
}

static void
Restore (void)
{
    UInt32 supportedDepths = 0;

    if (!PrefGetAppPreferencesV10 (PatienceFileCreator, PatienceVersionNum,
				   &game, sizeof (GameType)))
    {
        // preferences re-initialized
	game.playing = False;
	game.leftHanded = False;
	game.gameKind = -1;
        game.alwaysPack = true;
        game.centerTopSuit = true;

        // Depends on resource prefsUpActionList, etc.
        game.upActionIdx = buttonActionUndo;
        game.downActionIdx = buttonActionDeal;
        game.jogClickActionIdx = buttonActionHint;

        // preset volume
        if (romVersion >= sysVersion30)
            game.volume = PrefGetPreference (prefGameSoundVolume);
        else
        {
            UInt16 volume;
            // sndMaxAmp is 64 and max midi volume is 127,
            // so we store volume as a unsigned char
            SndGetDefaultVolume (0, 0, &volume);
            game.volume = volume;
        }

        // default for B&W is to use a pattern for the table
        supportedDepths = 1;
        if (romVersion >= sysVersion30 && ! game.forceBW)
        {
            WinScreenMode (winScreenModeGetSupportedDepths,
                    NULL, NULL, &supportedDepths, NULL);
        }

        if (supportedDepths & 0x01)
            game.useTablePattern = true;
    }
}

void
DeleteUndoDB (void)
{
    Err	    err;
    LocalID dbID;
    UInt16  cardNo;

    if (undoDB)
    {
	err = DmOpenDatabaseInfo (undoDB, &dbID, NULL, NULL, &cardNo, NULL);
	DmCloseDatabase (undoDB);
	undoDB = 0;
	if (!err)
	    DmDeleteDatabase (cardNo, dbID);
    }
    game.historyTop = 0;
    game.historyChunk.historyTop = 0;
}

static void
OpenUndoDB (void)
{
    Err	err;

    undoDB = DmOpenDatabaseByTypeCreator (PatienceUndoDBType,
            PatienceDBCreator, dmModeReadWrite);
    if (!undoDB)
    {
	err = DmCreateDatabase (0, "Patience History", PatienceDBCreator,
                PatienceUndoDBType, false);
	if (err)
	    return;
	undoDB = DmOpenDatabaseByTypeCreator (PatienceUndoDBType,
                PatienceDBCreator, dmModeReadWrite);
    }
}

// TODO put stats in a db so they're not reset each time system prefs are
// added
#if 0
static void
OpenStatsDB (void)
{
    Err	err;

    statsDB = DmOpenDatabaseByTypeCreator (PatienceStatsDBType,
            PatienceDBCreator, dmModeReadWrite);
    if (!statsDB)
    {
	err = DmCreateDatabase (0, "Patience Statistics", PatienceDBCreator,
                PatienceStatsDBType, false);
	if (err)
	    return;
	statsDB = DmOpenDatabaseByTypeCreator (PatienceStatsDBType,
                PatienceDBCreator, dmModeReadWrite);
        // TODO copy stats from preferences
    }
}
#endif

static void
ScreenInit (void)
{
    if (romVersion >= sysVersion30)
    {
        WinScreenMode (winScreenModeGet,
                NULL, NULL, &screen_depth, NULL );
    }
    else
    {
        screen_depth = 1;
    }

    save_depth = screen_depth;

    TryHRMode ();
    TryHandEraMode ();
    SetBestColorMode ();
    TrySilkMode ();
    TryDIAMode ();
}

static void
ScreenFini (void)
{
    UInt16 cardNo;
    LocalID dbID;

    SysCurAppDatabase (&cardNo, &dbID);

    if (silkMode)
    {
        // unregister sysNotifyDisplayChangeEvent notification
        SysNotifyUnregister (cardNo, dbID,
                sysNotifyDisplayChangeEvent, sysNotifyNormalPriority);
        SilkLibDisableResize (silkRefNum);
        SilkLibClose (silkRefNum);
    }
    if (diaMode)
    {
        // unregister sysNotifyDisplayResizedEvent notification
        SysNotifyUnregister (cardNo, dbID,
                sysNotifyDisplayResizedEvent, sysNotifyNormalPriority);
    }
    if (hrMode==SonyHR)
    {
        HRWinScreenMode (hrRefNum, winScreenModeSetToDefaults,
                NULL, NULL, NULL, NULL );
        HRClose (hrRefNum);
    }
    if (romVersion >= sysVersion30)
    {
        WinScreenMode (winScreenModeSet,
                NULL, NULL, &save_depth, NULL );
    }
}

// Do not call this function before setting the screen mode
static void
TrySilkMode (void)
{
    Err error = 0;
    UInt16 cardNo;
    LocalID dbID;

    // detect Sony silkscreen manager
    if ((error = SysLibFind (sonySysLibNameSilk, &silkRefNum))) {
        if (error == sysErrLibNotFound) {
            /* couldn't find lib, load it */
            error = SysLibLoad ('libr', sonySysFileCSilkLib, &silkRefNum);
        }
    }
    if (!error) {
        /* Version 1 is installed, only resize is available */
        if (SilkLibOpen (silkRefNum) == errNone) {
            SilkLibEnableResize (silkRefNum);
            // register for sysNotifyDisplayChangeEvent notifications;
            // a sysNotifyDisplayChangeEvent notification is sent when the
            // user resizes the silkscreen area
            // NOTE: these notifications must be registered AFTER changing
            //       the screen resolution/bit-depth
            if ((error = SysCurAppDatabase (&cardNo, &dbID)) != errNone
                || (error = SysNotifyRegister (cardNo, dbID,
                                sysNotifyDisplayChangeEvent, NULL,
                                sysNotifyNormalPriority, NULL)) != errNone)
            {
                ErrNonFatalDisplayIf(error != errNone,
                        "failed to register for sysNotifyDisplayChangeEvent");
                SilkLibClose(silkRefNum);
            }

            // Remember window size on exit and restore on re-entry
            if (game.silkStartNormal)
                SilkLibResizeDispWin (silkRefNum, silkResizeNormal);
            else
                SilkLibResizeDispWin (silkRefNum, silkResizeToStatus);

            silkMode = true;
        }
    }
}

// Look for PalmOS 5.3 Dynamic Input Area support
static void
TryDIAMode (void)
{
    Err         error = 0;
    UInt32      version = 0;
    UInt16 cardNo;
    LocalID dbID;

    // Look for the Pen Input Manager feature
    error = FtrGet(pinCreator, pinFtrAPIVersion, &version);
    if (!error && version)
        diaMode = true;
    else
        return;

    // register for sysNotifyDisplayResizedEvent notifications;
    // a sysNotifyDisplayResizedEvent notification is sent when the
    // user resizes the silkscreen area
    if ((error = SysCurAppDatabase (&cardNo, &dbID)) != errNone
        || (error = SysNotifyRegister (cardNo, dbID,
                        sysNotifyDisplayResizedEvent, NULL,
                        sysNotifyNormalPriority, NULL)) != errNone)
    {
        ErrNonFatalDisplayIf(error != errNone,
                "failed to register for sysNotifyDisplayResizedEvent");
    }

    // Remember window size on exit and restore on re-entry
    if (game.silkStartNormal)
        PINSetInputAreaState(pinInputAreaOpen);
    else
        PINSetInputAreaState(pinInputAreaClosed);
}

static void
TryHRMode (void)
{
    Err error = 0;
    MemHandle	handle = NULL;
    UInt32      version = 0;
    UInt32      attr = 0;

    error = FtrGet (sysFtrCreator, sysFtrNumWinVersion, &version);
    if (!error && version >= 4)
    {
        WinScreenGetAttribute(winScreenDensity, &attr);
        if (attr == kDensityDouble)
            hrMode = PalmOS5HR; //the screen is high density
        return;
    }

    // don't try to set high-res mode if the bitmap resources aren't present
    handle = DmGetResource (bitmapRsc, bitmapHRClubSuit);
    if (!handle)
    {
        // highres images not present, no highres mode
        return;
    }
    DmReleaseResource (handle);

    if ((error = SysLibFind(sonySysLibNameHR, &hrRefNum)))
    {
        if (error == sysErrLibNotFound)
        {
            /* couldn't find lib, load it */
            error = SysLibLoad('libr', sonySysFileCHRLib, &hrRefNum);
        }
    }
    if (!error )
    {
        /* Executes Open library. */
        error = HROpen(hrRefNum);
        if (error) {
            /* error processing */
        } else {
            UInt32 screen_width = hrWidth;
            UInt32 screen_height = hrHeight;
            error = HRWinScreenMode (hrRefNum, winScreenModeSet,
                    &screen_width, &screen_height, NULL, NULL );
            if (error != errNone) {
                /* Screen mode remains unchanged. */
            } else {
                /* high-resolution mode */
                hrMode = SonyHR;
            }
        }
    }
}

static void
TryHandEraMode (void)
{
    UInt32	version;
    MemHandle	handle = NULL;

    // don't try to set QVGA mode if the bitmap resources aren't present
    handle = DmGetResource (bitmapRsc, bitmapQVGAClubSuit);
    if (!handle)
    {
        // QVGA images not present, no highres mode
        return;
    }
    DmReleaseResource (handle);

    if (FtrGet(TRGSysFtrID, TRGVgaFtrNum, &version) == 0)
    {
        // Extension is present
        qvgaMode = true;
        VgaSetScreenMode(screenMode1To1, rotateModeNone);
    }
}

static void
SetBestColorMode (void)
{
    UInt32 supportedDepths = 0;

    if (romVersion >= sysVersion30 && ! game.forceBW)
    {
        WinScreenMode (winScreenModeGetSupportsColor,
                NULL, NULL, NULL, &supports_color);
        WinScreenMode (winScreenModeGetSupportedDepths,
                NULL, NULL, &supportedDepths, NULL);
    }
    else
    {
        supportedDepths = 1;
        supports_color = false;
        game.forceBW = true;
    }

    // Note: No need for more than 256 color mode

    if (supportedDepths & 0x80 && supports_color)
    {
        // 256 color mode
        screen_depth = 8;
    }
    else if (supportedDepths & 0x08)
    {
        // 16 color mode
        screen_depth = 4;
    }
    else if (supportedDepths & 0x02)
    {
        // 4 color mode
        screen_depth = 2;
    }
    else
    {
        // B&W
        screen_depth = 1;
    }

    if (romVersion >= sysVersion30)
    {
        // --doesn't work
        //WinScreenMode (winScreenModeSet,
        //        NULL, NULL, &screen_depth, &supports_color );
        WinScreenMode (winScreenModeSet,
                NULL, NULL, &screen_depth, NULL );
    }

    if (romVersion >= sysVersion35)
    {
        RGBColorType white  = COLOR_white;
        const RGBColorType black  = COLOR_black;
        const RGBColorType green  = COLOR_green;
        const RGBColorType maroon = COLOR_maroon;
        const RGBColorType grey   = COLOR_grey;
        const RGBColorType silver = COLOR_silver;
        const RGBColorType gold   = COLOR_gold;

        tableIndex = WinRGBToIndex(&green);
        cardBackIndex = WinRGBToIndex(&maroon);
        blackIndex = WinRGBToIndex(&black);
        packIndex = WinRGBToIndex(&silver);
        if (screen_depth < 8)
            hintIndex = WinRGBToIndex(&grey);
        else
            hintIndex = WinRGBToIndex(&gold);
        // Create a non-transparent white
        // (the default transparency index is 0, which is white)
        if (WinPalette (winPaletteSet, 254, 1, &white) == errNone)
            whiteIndex = 254;
        else
            whiteIndex = 0;
    }
}

static void
StartApplication (void)
{
    // what OS are we on?
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    sysVersion30 = sysMakeROMVersion(3,0,0,0,0);
    sysVersion33 = sysMakeROMVersion(3,3,0,0,0);
    sysVersion35 = sysMakeROMVersion(3,5,0,0,0);
    sysVersion40 = sysMakeROMVersion(4,0,0,0,0);
    sysVersion50 = sysMakeROMVersion(5,0,0,0,0);
    // still supports 2.0!

    mainFrm = NULL;
    MemSet (&game, sizeof (GameType), '\0');
    OpenUndoDB ();
    //OpenStatsDB ();
    Restore ();
    ScreenInit ();
    CardLabelInit ();
    WinResetClip ();
    game.nhint = 0;
    game.queryMode = False;
    game.srcStack = NullIndex;
    game.srcCard = NullIndex;
}

#define abs(x)	((x) >= 0 ? (x) : -(x))

static void
CheckTracking (PatI x, PatI y)
{
    if (game.srcStack != NullIndex &&
	game.srcCard != NullIndex)
    {
	if (TimGetTicks() >= downTime + TRACK_TIMEOUT ||
	    abs (x - downX) > TRACK_BOX || abs (y - downY) > TRACK_BOX)
	{
	    tracking = CardStartAnimate (x, y, game.srcStack, game.srcCard);
	    awaitingTracking = false;
	}
    }
}

static void
CheckQuery (PatI x, PatI y)
{
    CardStackIndex  stack;
    CardIndex	    card;
    PatI	    c_x, c_y;

    stack = NullIndex;
    if (!CardHit (x, y, &stack, &card, &c_x, &c_y))
	return;

    if (card == NullIndex)
	return;

    if (stack == game.srcStack && card == game.srcCard)
	return;

    game.srcStack = stack;
    game.srcCard = card;
    tracking = CardStartAnimate (c_x, c_y, stack, card);
}

static Boolean
MainFormHandleEvent(EventPtr event)
{
    Boolean	    wasPlaying;
    Boolean	    handled = false;
    Boolean	    shouldTrack = false;
    CardStackIndex  dstStack;
    CardIndex	    dstCard;
    VgaScreenModeType screenMode;
    VgaRotateModeType rotateMode;
    PatS	    gameSeed;

    wasPlaying = game.playing;
    switch (event->eType) {
    case penDownEvent:
	if (0 <= event->screenX && event->screenX < form_width &&
	    0 <= event->screenY && event->screenY < form_height &&
	    game.playing)
	{
	    pressed = true;
	    handled = true;
	    game.srcStack = NullIndex;
	    downX = event->screenX;
	    downY = event->screenY;
	    downTime = TimGetTicks ();
	    if (game.queryMode)
	    {
		CheckQuery (event->screenX, event->screenY);
		shouldTrack = tracking;
	    }
	    else
	    {
		(void) CardHit (event->screenX, event->screenY,
				&game.srcStack, &game.srcCard,
				0, 0);
		awaitingTracking = true;
	    }
	}
	break;
    case penMoveEvent:
	handled = pressed;
	if (game.playing)
	{
	    if (game.queryMode)
	    {
		downX = event->screenX;
		downY = event->screenY;
		CheckQuery (event->screenX, event->screenY);
	    }
	    else
	    {
		if (tracking)
		    CardMoveAnimate (event->screenX, event->screenY);
		else
		    CheckTracking (event->screenX, event->screenY);
	    }
	    shouldTrack = tracking;
	}
	break;
    case penUpEvent:
	handled = pressed;
	pressed = false;
	awaitingTracking = false;
	if (tracking)
	{
	    CardStopAnimate ();
	    tracking = False;
	}
	if (game.playing && !game.queryMode)
	{
	    if (game.srcStack != NullIndex &&
		CardHit (
			 event->data.penUp.end.x, event->data.penUp.end.y,
			 &dstStack, &dstCard,
			 0, 0))
	    {
		CardPlay (games[game.gameKind],
			  game.srcStack, dstStack, game.srcCard);
	    }
	}
	game.srcStack = NullIndex;
	break;
    case frmUpdateEvent:
	if (event->data.frmUpdate.formID == mainForm)
	{
            // If we need to distinguish, our updates are
            // event->data.frmUpdate.updateCode == 5, the
            // system sends frmRedrawUpdateCode
            if (window_resized)
            {
                WinResetClip ();
                SizeFormWindow (mainFrm, NULL, NULL);
                window_resized = false;
            }
            Redisplay ();
            needRedisplay = false;
            handled = true;
	}
	break;
    // PalmOS dynamic input area
    case winDisplayChangedEvent:
        window_resized = true;
        SizeFormWindow(FrmGetActiveForm(), NULL, NULL);
        handled = true;
        break;
    // Sony silkscreen manager
    case winEnterEvent :
        // if we are returning from another window (such as a dialog
        // box), check to see if the Main Form needs updating as
        // a result of resizing the silkscreen area
        //
        // (we must check for winEnter events because modal Forms with
        // the save-behind bit enabled may not generate frmUpdate
        // events)
        //
        // pdc-This check is needed for main form since other forms draw
        // over the main form and clear the window_resized flag.  When they
        // go away, we need to detect if the main window size has changed.
        if (event->data.winEnter.enterWindow == FrmGetWindowHandle (mainFrm))
        {
            Int16 heightDelta;

            if (diaMode) {
                // indicate window can be resized
                FrmSetDIAPolicyAttr(FrmGetActiveForm (), frmDIAPolicyCustom);
                PINSetInputTriggerState(pinInputTriggerEnabled);
                PINSetInputAreaState(pinInputAreaUser);
            }

            SizeFormWindow (mainFrm, NULL, &heightDelta);
            if (heightDelta != 0)
                FrmUpdateForm (mainForm, frmRedrawUpdateCode);
        }
        //if (event->data.winEnter.enterWindow ==
        //        FrmGetWindowHandle (FrmGetActiveForm ()) && window_resized)
        //    FrmUpdateForm (FrmGetActiveFormID (), frmRedrawUpdateCode);
        break;
    case frmOpenEvent:
	if (event->data.frmOpen.formID == mainForm)
	{
            SizeFormWindow (mainFrm, NULL, NULL);
	    FrmDrawForm (mainFrm);
	    if (game.playing)
	    {
                SetSizes ();
		needRedisplay = true;
	    }
	    else
	    {
		Redisplay ();
		SelectGame (false, false);
	    }
	}
	break;
    // HandEra
    case displayExtentChangedEvent:
        SizeFormWindow(FrmGetActiveForm(), NULL, NULL);
        handled = true;
        break;
    // HandEra
    case menuOpenEvent:
        if (qvgaMode)
            MenuAddItem(menuPatienceAlwaysAuto, menuRotate, 0, "Rotate...");
        handled = true;
        break;
    case menuCmdBarOpenEvent:
        // Prevent the command toolbar from being displayed.  Alternatively,
        // could call MenuEraseStatus before executing a menu command that
        // might draw on the lower portion of the window.
        handled = true;
        break;
    case menuEvent:
	switch (event->data.menu.itemID) {
	case menuPatienceHint:
	    if (game.playing)
		CardHint (games[game.gameKind]);
	    break;
	case menuPatienceAuto:
	    if (game.playing)
	    {
		CardAutoPlay (games[game.gameKind]);
		CardRedisplay ();
	    }
	    break;
	case menuPatienceUndo:
	    if (game.playing)
		CardUndo ();
	    break;
	case menuPatienceSameGame:
	    if (game.playing)
	        SelectGame (false, true);
	    break;
	case menuPatienceRestart:
	    if (game.playing)
		RestartGame ();
	    break;
	case menuPatienceNew:
	    SelectGame (false, false);
	    break;
        case menuPatienceStats:
            StatisticsForm ();
            break;
        case menuPatienceGameInfo:
            gameSeed = GameInfoForm ();
            if (gameSeed != -1)
            {
                game.gameSeed = gameSeed;
                // TODO need select game given the seed
                // TODO should add game number to new screen too
                if (game.playing)
                    RestartGame ();
            }
            break;
	case menuPatienceAlwaysAuto:
	    if (game.playing)
		SetAutoMode (!game.alwaysAuto);
	    break;
	case menuPatienceExamine:
	    if (game.playing)
		SetQueryMode (!game.queryMode);
	    break;
	case menuPatienceLeftHanded:
	    SetHanded (!game.leftHanded, true);
	    break;
        case menuPatiencePrefs:
            Preferences();
            break;
	case menuHelpControls:
            Rules (controlsHelp);
	    break;
	case menuHelpRules:
	    Rules (game.gameKind + rulesFirst);
	    break;
	case menuHelpGetInfo:
	    Help (infoForm);
	    break;
        // HandEra
        case menuRotate:
            VgaGetScreenMode (&screenMode, &rotateMode);
            VgaSetScreenMode (screenMode, VgaRotateSelect (rotateMode));
            SizeFormWindow (FrmGetActiveForm(), NULL, NULL);
            break;
	}
	handled = true;
	break;
    case keyDownEvent:
	switch (event->data.keyDown.chr) {
	case 'c':
	case 'C':
            Rules (controlsHelp);
	    handled = true;
	    break;
	case 'r':
	case 'R':
	    Rules (game.gameKind + rulesFirst);
	    handled = true;
	    break;
	case 'i':
	case 'I':
	    Help (infoForm);
	    handled = true;
	    break;
	case 'h':
	case 'H':
	case ' ':
	    if (game.playing)
		CardHint (games[game.gameKind]);
	    handled = true;
	    break;
	case 'a':
	case 'A':
	    if (game.playing)
	    {
		CardAutoPlay (games[game.gameKind]);
		CardRedisplay ();
	    }
	    handled = true;
	    break;
	case 'u':
	case 'U':
	case '\b':
	    if (game.playing)
		CardUndo ();
	    handled = true;
	    break;
	case 's':
	case 'S':
	    RestartGame ();
	    handled = true;
	    break;
        case 't':
        case 'T':
            StatisticsForm ();
            handled = true;
            break;
	case 'm':
	case 'M':
	    SelectGame (false, true);
	    handled = true;
	    break;
#if DEBUG
	case 'q':   // test new game form (in stats.c)
	case 'Q':
	    NewGameForm (false);
	    handled = true;
	    break;
#endif
	case 'n':
	case 'N':
	    SelectGame (false, false);
	    handled = true;
	    break;
	case 'e':
	case 'E':
	    SetQueryMode (!game.queryMode);
	    if (tracking && !game.queryMode)
	    {
		CheckTracking (downX, downY);
		shouldTrack = tracking;
		break;
	    }
	    handled = true;
	    break;
	case 'l':
	case 'L':
	    SetHanded (!game.leftHanded, true);
	    handled = true;
	    break;
	case 'p':
	case 'P':
	    PlayRandomSong (True);
	    handled = true;
	    break;
#if DEBUG
        case 'f': // test entertain function
        case 'F':
            GameWonEntertain ();
	    handled = true;
	    break;
#endif
        case 'd':
        case 'D':
            {
            CardStackType *sp = CardStackPtr (CardStackDeck);
            CardPlay (games[game.gameKind],
                      CardStackDeck, CardStackDeck, sp->first);
	    handled = true;
            }
            break;
        // configurable buttons
	case vchrPageUp:
            HandleAction (game.upActionIdx);
	    handled = true;
	    break;
	case vchrPageDown:
            HandleAction (game.downActionIdx);
	    handled = true;
	    break;
	case chrCarriageReturn: // HandEra Jog Click
	case vchrJogRelease:    // Sony Jog Click
            HandleAction (game.jogClickActionIdx);
	    handled = true;
	    break;
	}
        // Palm 5-Way controller support (Tungsten/etc)
        //TODO also NavDirectionPressed(event*, Left/Right)
        if (!handled && NavSelectPressed(event))
        {
            HandleAction (game.jogClickActionIdx);
            handled = true;
        }
	break;
    default:
    }
    if (tracking && !shouldTrack)
    {
	CardStopAnimate ();
	tracking = False;
    }

    if (wasPlaying && !game.playing)
    {
        GameWonEntertain ();
    	SelectGame (true, false);
    }

    return(handled);
}

static void
HandleAction (ButtonAction action)
{
    Boolean	    shouldTrack = false;

    switch (action)
    {
    case buttonActionUndo:
        if (game.playing)
            CardUndo ();
        break;
    case buttonActionHint:
        if (game.playing)
            CardHint (games[game.gameKind]);
        break;
    case buttonActionExamine:
        SetQueryMode (!game.queryMode);
        if (tracking && !game.queryMode)
        {
            CheckTracking (downX, downY);
            shouldTrack = tracking;
        }
        break;
    case buttonActionDeal:
        {
        CardStackType *sp = CardStackPtr (CardStackDeck);
        CardPlay (games[game.gameKind],
                  CardStackDeck, CardStackDeck, sp->first);
        }
        break;
    }
    if (tracking && !shouldTrack)
    {
	CardStopAnimate ();
	tracking = False;
    }
}

static Boolean ApplicationHandleEvent (EventPtr event)
{
    UInt16 formId;
    FormPtr frm;

    if (event->eType == frmLoadEvent)
    {
	// Load the form resource.
	formId = event->data.frmLoad.formID;
	frm = FrmInitForm (formId);
	if (frm)
	{
	    FrmSetActiveForm (frm);

	    switch (formId) {
	      case mainForm:
		mainFrm = frm;
		FrmSetEventHandler (frm, MainFormHandleEvent);
                if (diaMode)
                {
                    // indicate window can be resized
                    FrmSetDIAPolicyAttr(frm, frmDIAPolicyCustom);
                    PINSetInputTriggerState(pinInputTriggerEnabled);
                    PINSetInputAreaState(pinInputAreaUser);
                    // min preferred and max sizes, height and width
                    // (440 = 10 cards x 44 max card width)
                    WinSetConstraintsSize (FrmGetWindowHandle (frm),
                            120, 640, 640, 160, 440, 640);
                }
		break;
              case rulesForm:
                if (diaMode)
                {
                    // indicate window can be resized
                    FrmSetDIAPolicyAttr(frm, frmDIAPolicyCustom);
                    PINSetInputTriggerState(pinInputTriggerEnabled);
                    PINSetInputAreaState(pinInputAreaUser);
                    // min preferred and max sizes, height and width
                    WinSetConstraintsSize (FrmGetWindowHandle (frm),
                            80, 480, 640, 80, 156, 320);
                }
                break;
              case statsForm:
              case newGameForm:
                if (diaMode)
                {
                    // indicate window can be resized
                    FrmSetDIAPolicyAttr(frm, frmDIAPolicyCustom);
                    PINSetInputTriggerState(pinInputTriggerEnabled);
                    PINSetInputAreaState(pinInputAreaUser);
                    // min preferred and max sizes, height and width
                    WinSetConstraintsSize (FrmGetWindowHandle (frm),
                            120, 480, 640, 160, 160, 160);
                }
		break;
	    }
	}
	return true;
    }
    return false;
}

static void
EventLoop(void)
{
    EventType	event;
    Int32	timeout, now;
    UInt16	error;

    do
    {
	if (needRedisplay)
	    timeout = 0;
	else if (awaitingTracking && !tracking)
	{
	    now = TimGetTicks();
	    timeout = downTime + TRACK_TIMEOUT;
	    if (timeout < now)
		timeout = 0;
	    else
		timeout = timeout - now;
	}
	else if (game.nhint > 0)
	    timeout = TRACK_TIMEOUT/2;
	else
	    timeout = evtWaitForever;
	// Get the next available event.
	EvtGetEvent(&event, timeout);

	if (needRedisplay && event.eType == nilEvent)
	    //FrmUpdateForm (mainForm, frmRedrawUpdateCode);
	    FrmUpdateForm (mainForm, 5);

	if (awaitingTracking && !tracking)
	    CheckTracking (downX, downY);

	//if (game.nhint > 0 && event.eType != penUpEvent)

        // Unshowhint needs to be called before any drawing;
        // check the event type and make a decision.  This is
        // also checked before Redisplay, CardRedisplay,
        // and drawing new hints.
	if (game.nhint > 0 && (
                TimGetTicks() > hintDrawTicks + HINT_TIMEOUT ||
                event.eType == penDownEvent ||
                event.eType == frmOpenEvent ||
                event.eType == frmUpdateEvent ||
                event.eType == menuOpenEvent))
	    CardUnshowHints ();

	if (event.eType == nilEvent)
	    continue;

        // TODO support "hard" keys
        //switch (event.eType) {
        //case keyDownEvent:
        //    switch (event.data.keyDown.chr) {
        //    case vchrHard1:
        //        if (game.playing)
        //            CardUndo ();
        //        continue;
        //        break;
        //    }
        //default:
        //}

	/* Give the system a chance to handle the event. */
	if (! SysHandleEvent (&event))
	{
	    if (!MenuHandleEvent (0, &event, &error))
	    {
		/* Give the application a chance to handle the event. */
		if (!ApplicationHandleEvent (&event))
		{
		    /* Let the form object provide default handling of the event. */
		    FrmDispatchEvent (&event);
		}
	    }
	}
    }
    while (event.eType != appStopEvent);
    // ** SPECIAL NOTE **
    // In order for the Emulator to exit
    // cleanly when the File|Quit menu option is
    // selected, the running application
    // must exit.  The emulator will generate an
    // "appStopEvent" when Quit is selected.
}

static void
StopApplication(void)
{
    if (tracking)
    {
	CardStopAnimate ();
	tracking = False;
    }

    // Remember window size on exit and restore on re-entry
    game.silkStartNormal = (form_height == form_width);

    Save ();
    CardLabelFini ();

    // P11. Close all open forms to allow their frmCloseEvent handlers
    // to execute.  An appStopEvent doesn't send frmCloseEvents.
    // FrmCloseAllForms will send all opened forms a frmCloseEvent.
    FrmCloseAllForms ();
    if (undoDB)
    {
	DmCloseDatabase (undoDB);
	undoDB = 0;
    }
    if (statsDB)
    {
	DmCloseDatabase (statsDB);
	statsDB = 0;
    }

    ScreenFini ();
}

#if 0
#define DIAG(s)	printf(s); SysTaskDelay (sysTicksPerSecond/4);
#else
#define DIAG(s)
#endif

UInt32
PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
    switch (cmd) {
      case sysAppLaunchCmdNormalLaunch:
	DIAG ("StartApplication\n");
	// Set up initial form.
	StartApplication();

	DIAG ("FrmGotoForm\n");
	FrmGotoForm (mainForm);

	// Start up the event loop.
	DIAG ("EventLoop\n");
	EventLoop();

	// Clean up before exiting
	DIAG ("StopApplication\n");
	StopApplication();
	DIAG ("Patience done\n");
        break;

      case sysAppLaunchCmdNotify:
      {
        SysNotifyParamType* notifyParams = (SysNotifyParamType*) cmdPBP;
        ErrFatalDisplayIf(cmdPBP == NULL,
                "Patience: PilotMain: null notification parameters");

        // PalmOS dynamic input area approach
        if (notifyParams->notifyType == sysNotifyDisplayResizedEvent) {
            // Pen Input Manager 1.0 doesn't post winDisplayChangedEvent
            // but 1.1 does.  Add winDisplayChangedEvent for compatibility
            EventType e;
            MemSet (&e, sizeof(EventType), 0);
            e.eType = winDisplayChangedEvent;
            EvtAddUniqueEventToQueue (&e, 0, true);
        }

        // Sony silk manager approach
        // Notification Manager sysNotifyDisplayChangeEvent
        //
        // The sysNotifyDisplayChangeEvent is broadcast whenever the display
        // mode changes.  That is, either the color table has been set to
        // use a specific palette using the WinPalette function or the bit
        // depth has changed using the WinScreenMode function.
        if (notifyParams->notifyType == sysNotifyDisplayChangeEvent) {
            Boolean appIsActive = (launchFlags & sysAppLaunchFlagSubCall);
            if (appIsActive)
            {
                SysNotifyDisplayChangeDetailsType *details =
                        (SysNotifyDisplayChangeDetailsType*)
                        notifyParams->notifyDetailsP;

                // When the silkscreen is minimized, oldDepth == newDepth
                if (details->oldDepth == details->newDepth)
                {
                    UInt16 activeFormID = FrmGetActiveFormID ();
                    //TODO still not quite right (but seems to be good enough)
                    // Only refresh if the window is what we think it is
                    // (could also be find, alarm, etc)

                    // problem is forms get started over main form, and then
                    // there's no way to alert main form the window size has
                    // changed.

                    // these forms support resize
                    if (activeFormID == mainForm ||
                            activeFormID == rulesForm ||
                            activeFormID == statsForm ||
                            activeFormID == newGameForm)
                    {
                        FrmUpdateForm (activeFormID, frmRedrawUpdateCode);
                    }
                    window_resized = true;
                }

                notifyParams->handled = true;
            }
        }
        break;
      }

      default:
        break;
    }

    return 0;
}
